from typing import Dict, Optional

import requests
from requests.sessions import Session

from caplena.http.http_client import HttpClient, HttpMethod, HttpRetry
from caplena.http.http_response import HttpResponse


class RequestsHttpClient(HttpClient):
    @property
    def identifier(self) -> str:
        return f"requests({requests.__version__})"

    def __init__(
        self,
        *,
        timeout: int = HttpClient.DEFAULT_TIMEOUT,
        retry: HttpRetry = HttpClient.DEFAULT_RETRY,
        session: Optional[Session] = None,
    ):
        super().__init__(timeout=timeout, retry=retry)
        self.session = session if session is not None else Session()

    def request_raw(
        self,
        uri: str,
        *,
        method: HttpMethod,
        timeout: int,
        headers: Dict[str, str],
        data: Optional[str] = None,
    ) -> HttpResponse:
        response = self.session.request(
            url=uri,
            method=method.method,
            data=data,
            headers=headers,
            timeout=timeout,
        )
        if response.status_code > 500:
            response.raise_for_status()

        # note: we only support utf-8 encodings
        if response.encoding != "utf-8":
            raise ValueError(
                f"Received a response with an unsupported encoding scheme (encoding='{response.encoding}')."
            )

        return HttpResponse(
            status_code=response.status_code,
            reason=response.reason,
            text=response.text,
            headers=dict(response.headers),
        )



import unittest
from datetime import datetime
from typing import Any, Dict, List, Optional

from typing_extensions import Literal

from caplena.constants import NOT_SET
from caplena.endpoints.base_endpoint import BaseController, BaseObject, BaseResource
from caplena.helpers import Helpers
from caplena.list import CaplenaList
from tests.common import common_config


class SomeController(BaseController):
    pass


class SomeObject(BaseResource[SomeController]):
    class SomeNested(BaseObject[SomeController]):
        class Metadata(BaseObject[SomeController]):
            __fields__ = {"some_other_field", "reviewed_count"}
            __mutable__ = {"reviewed_count"}
            some_other_field: int
            reviewed_count: int

        __fields__ = {"ref", "name", "type", "metadata"}
        __mutable__ = {"ref"}
        ref: str
        name: str
        type: Literal["numerical", "boolean", "text", "date", "any", "text_to_analyze"]
        metadata: Metadata

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "SomeObject.SomeNested":
            obj["metadata"] = cls.Metadata.parse_obj(obj["metadata"])
            return super().parse_obj(obj)

    __fields__ = {
        "name",
        "tags",
        "upload_status",
        "columns",
        "created",
        "last_modified",
        "translation_status",
        "nested",
    }
    __mutable__ = {"name", "tags", "created", "translation_status", "nested"}

    name: str
    tags: List[str]
    upload_status: Literal["pending", "in_progress", "succeeded", "failed"]
    columns: CaplenaList[SomeNested]
    created: datetime
    last_modified: datetime
    translation_status: Optional[str]
    nested: SomeNested

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "SomeObject":
        obj["columns"] = CaplenaList(
            values=[cls.SomeNested.parse_obj(column) for column in obj["columns"]]
        )
        obj["nested"] = cls.SomeNested.parse_obj(obj["nested"])
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)


class CustomizedObject(SomeObject):
    class SomeNested(SomeObject.SomeNested):
        def modified_dict(self) -> Any:
            modified: Any = super().modified_dict()
            if modified == NOT_SET:
                modified = {}

            modified["ref"] = self.ref
            modified["type"] = self.type
            return modified


class BaseObjectTests(unittest.TestCase):
    controller = SomeController(config=common_config)

    def build_object_dict(self) -> Dict[str, Any]:
        return {
            "id": "id_object",
            "name": "Hello",
            "tags": ["tag1", "tag2"],
            "upload_status": "pending",
            "columns": [
                {
                    "ref": "column_1",
                    "name": "Column 1",
                    "type": "numerical",
                    "metadata": {"some_other_field": 12, "reviewed_count": 42},
                },
                {
                    "ref": "column_2",
                    "name": "Column 2",
                    "type": "date",
                    "metadata": {"some_other_field": 400, "reviewed_count": 40000},
                },
            ],
            "created": "2022-03-14T08:18:38.910Z",
            "last_modified": "2021-03-14T08:18:38.905Z",
            "translation_status": None,
            "nested": {
                "ref": "some_ref",
                "name": "nested name",
                "type": "numerical",
                "metadata": {"some_other_field": 12, "reviewed_count": 42},
            },
        }

    def build_obj(self, *, obj_exists: bool = False) -> SomeObject:
        return SomeObject.build_obj(
            self.build_object_dict(), controller=self.controller, obj_exists=obj_exists
        )

    def build_customized(self, *, obj_exists: bool = False) -> CustomizedObject:
        return CustomizedObject.build_obj(
            self.build_object_dict(), controller=self.controller, obj_exists=obj_exists
        )

    # ----- General Functionality ----- #

    def test_parsing_objects_succeeds(self) -> None:
        obj = SomeObject.parse_obj(self.build_object_dict())

        self.assertEqual(obj.id, "id_object")
        self.assertEqual(obj.name, "Hello")
        self.assertEqual(obj.tags, ["tag1", "tag2"])
        self.assertEqual(obj.created, Helpers.from_rfc3339_datetime("2022-03-14T08:18:38.910Z"))
        self.assertEqual(obj.translation_status, None)
        self.assertEqual(obj.nested.ref, "some_ref")
        self.assertEqual(obj._previous, {})  # pyright: reportPrivateUsage=false

    def test_building_object_succeeds(self) -> None:
        # test: test building object
        obj = SomeObject.build_obj(
            self.build_object_dict(), controller=self.controller, obj_exists=False
        )
        self.assertEqual(obj.id, "id_object")
        self.assertEqual(obj.name, "Hello")
        self.assertEqual(obj.tags, ["tag1", "tag2"])
        self.assertEqual(obj.created, Helpers.from_rfc3339_datetime("2022-03-14T08:18:38.910Z"))
        self.assertEqual(obj.translation_status, None)
        self.assertEqual(obj.nested.ref, "some_ref")
        self.assertEqual(obj._previous, {})  # pyright: reportPrivateUsage=false

        # test: building object from api response
        obj = SomeObject.build_obj(
            self.build_object_dict(), controller=self.controller, obj_exists=True
        )
        self.assertEqual(obj.id, "id_object")
        self.assertEqual(obj.name, "Hello")
        self.assertEqual(obj.tags, ["tag1", "tag2"])
        self.assertEqual(obj.created, Helpers.from_rfc3339_datetime("2022-03-14T08:18:38.910Z"))
        self.assertEqual(obj.translation_status, None)
        self.assertEqual(obj.nested.ref, "some_ref")
        self.assertDictEqual(obj._previous, obj._attrs)  # pyright: reportPrivateUsage=false

        # test: make sure objects are actually deep copied
        obj._previous["columns"][0].ref = "column_invalid"
        self.assertEqual(obj._previous["columns"][0].ref, "column_invalid")
        self.assertEqual(obj._attrs["columns"][0].ref, "column_1")

    def test_object_equivalence_fails(self) -> None:
        first = self.build_obj()

        # test: simple property differs
        second = self.build_obj()
        second.name = "This property differs"
        self.assertNotEqual(second, first)

        # test: simple nested property differs
        second = self.build_obj()
        second.nested.ref = "invalid_ref"
        self.assertNotEqual(second, first)

        # test: doubly nested property differs
        second = self.build_obj()
        second.nested.metadata.reviewed_count = 10000
        self.assertNotEqual(second, first)

        # test: simple list differs
        second = self.build_obj()
        second.tags = ["diff"]
        self.assertNotEqual(second, first)

        # test: simple nested list differs
        second = self.build_obj()
        second.columns[0].ref = "invalid_ref"
        self.assertNotEqual(second, first)

        # test: complex nested list differs
        second = self.build_obj()
        second.columns[0].metadata.reviewed_count = 10000
        self.assertNotEqual(second, first)

    def test_object_equivalence_succeeds(self) -> None:
        first = self.build_obj()

        # test: simple equivalence test
        second = self.build_obj()
        self.assertEqual(first, second)

        # test: changing properties works
        second.name = "Not the same anymore."
        self.assertNotEqual(first, second)
        second.nested.ref = "still not the same."
        self.assertNotEqual(first, second)
        second.name = "Hello"
        self.assertNotEqual(first, second)
        second.nested.ref = "some_ref"
        self.assertEqual(first, second)

    def test_modifying_forbidden_fields_fails(self) -> None:
        first = self.build_obj()

        # test: assigning simple forbidden field
        with self.assertRaisesRegex(AttributeError, "upload_status"):
            first.upload_status = "in_progress"

        # test: assigning simple nested field
        with self.assertRaisesRegex(AttributeError, "type"):
            first.nested.type = "text"

        # test: assigning doubly nested field
        with self.assertRaisesRegex(AttributeError, "some_other_field"):
            first.nested.metadata.some_other_field = 400

        # test: appending to list fails
        with self.assertRaisesRegex(ValueError, "Error appending item"):
            first.columns.append(first.columns[0])

        # test: removing from list fails
        with self.assertRaisesRegex(ValueError, "Error deleting item"):
            del first.columns[0]

        # test: replacing in list fails
        with self.assertRaisesRegex(ValueError, "Error setting item"):
            first.columns[0] = first.columns[1]

    def test_object_modified_dict_succeeds(self) -> None:
        first = self.build_obj(obj_exists=True)

        # test: nothing has changed
        modified = first.modified_dict()
        self.assertEqual(modified, NOT_SET)

        # test: simple property has changed
        first.name = "My New Name"
        modified = first.modified_dict()
        self.assertDictEqual(modified, {"name": "My New Name"})

        # test: simple list has changed
        first = self.build_obj(obj_exists=True)
        first.tags.append("new-tag")
        modified = first.modified_dict()
        self.assertDictEqual(modified, {"tags": ["tag1", "tag2", "new-tag"]})

        # test: simple date has changed
        now = datetime.now()
        first = self.build_obj(obj_exists=True)
        first.created = now
        modified = first.modified_dict()
        self.assertDictEqual(modified, {"created": now})

        # test: simple nested field has changed
        first = self.build_obj(obj_exists=True)
        first.nested.ref = "new_ref"
        modified = first.modified_dict()
        self.assertDictEqual(modified, {"nested": {"ref": "new_ref"}})

        # test: many fields have changed
        first = self.build_obj(obj_exists=True)
        now = datetime(year=2020, month=10, day=10, hour=0)
        first.name = "new name"
        first.tags = ["new-tags"]
        first.created = now
        first.nested.ref = "new_ref"
        first.nested.metadata.reviewed_count = 400
        first.columns[0].ref = "obj_ref"
        first.columns[1].metadata.reviewed_count = 200
        modified = first.modified_dict()
        self.assertDictEqual(
            modified,
            {
                "name": "new name",
                "tags": ["new-tags"],
                "created": now,
                "nested": {"ref": "new_ref", "metadata": {"reviewed_count": 400}},
                "columns": [{"ref": "obj_ref"}, {"metadata": {"reviewed_count": 200}}],
            },
        )

        # test: only partial nested list fields have changed
        first = self.build_obj(obj_exists=True)
        first.columns[0].ref = "obj_ref"
        first.columns[0].metadata.reviewed_count = 80000
        modified = first.modified_dict()
        self.assertDictEqual(
            modified,
            {"columns": [{"ref": "obj_ref", "metadata": {"reviewed_count": 80000}}]},
        )

    def test_customizing_modified_dict_succeeds(self) -> None:
        # test: nothing has changed
        first = self.build_customized(obj_exists=True)
        modified = first.modified_dict()
        self.assertEqual(modified, NOT_SET)

        # test: one entry has changed
        first = self.build_customized(obj_exists=True)
        first.columns[0].ref = "obj_ref"
        first.columns[0].metadata.reviewed_count = 80000
        modified = first.modified_dict()
        self.assertDictEqual(
            modified,
            {
                "columns": [
                    {"ref": "obj_ref", "type": "numerical", "metadata": {"reviewed_count": 80000}},
                    {"ref": "column_2", "type": "date"},
                ]
            },
        )



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from caplena.api import ApiOrdering
from caplena.constants import LIST_PAGINATION_LIMIT, NOT_SET
from caplena.endpoints.base_endpoint import BaseController, BaseObject, BaseResource
from caplena.filters.projects_filter import ProjectsFilter, RowsFilter
from caplena.helpers import Helpers
from caplena.http.http_response import HttpResponse
from caplena.iterator import CaplenaIterator
from caplena.list import CaplenaList

# --- Controller --- #


class ProjectsController(BaseController):
    """The projects controller, encapsulating all project actions.

    :param config: The configuration object that a particular controller should use.
    """

    def create(
        self,
        *,
        name: str,
        language: str,
        columns: List[Dict[str, Any]],
        tags: Optional[List[str]] = NOT_SET,
        translation_engine: Optional[str] = NOT_SET,
        anonymize_pii: Optional[Dict[str, List[str]]] = NOT_SET,
    ) -> "ProjectDetail":
        """Creates a new project.

        :param name: Project name, as displayed in the user interface.
        :param language: Base language for this project.
        :param columns: Columns of a project define its schema. In a sense, every project column corresponds to
            exactly one column in an Excel sheet. The four column types numerical, date, boolean and text are
            auxiliary columns, meaning that they won't be analyzed, but they can be used to visualize your results.
            Please note that for columns of type `numerical`, its integer values must be between `-(2^53-1)` and
            `2^53-1`. For bigger numbers, please use a column of type `text`.
        :param tags: Tags assigned to this project. If omitted, no tags are assigned.
        :param translation_engine: Translation engine used to translate rows into the base language of this project.
            If omitted, no translation will be performed.
        :param anonymize_pii: settings to use for anonymization, visit the developer docs for more information.
        :raises caplena.api.ApiException: An API exception.
        """
        json = self.api.build_payload(
            name=name,
            language=language,
            columns=columns,
            tags=tags,
            translation_engine=translation_engine,
            anonymize_pii=anonymize_pii,
        )

        response = self.post(path="/projects", json=json)
        return self.build_response(response, resource=ProjectDetail)

    def retrieve(self, *, id: str) -> "ProjectDetail":
        """Retrieves a project you have previously created.

        :param id: The project identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        response = self.get(path="/projects/{id}", path_params={"id": id})
        return self.build_response(response, resource=ProjectDetail)

    def remove(self, *, id: str) -> None:
        """Removes a previously created project.

        :param id: The project identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        self.delete(path="/projects/{id}", path_params={"id": id})

    def list(
        self,
        *,
        order_by: ApiOrdering = ApiOrdering.desc("last_modified"),
        limit: Optional[int] = None,
        filter: Optional[ProjectsFilter] = None,
    ) -> "CaplenaIterator[ListedProject]":
        """Returns an iterator of all projects you have previously created. By default, the projects are returned
        in sorted order, with the most recently modified project appearing first.

        :param order_by: Column on which the results should be ordered on. Defaults to :code:`desc:last_modified`.
        :type order_by: ApiOrdering
        :param limit: Number of results returned per page. If unspecified, will return all results.
        :param filter: Filters to apply to this request. If omitted, no filters are applied.
        :raises caplena.api.ApiException: An API exception.
        """

        def fetcher(page: int) -> HttpResponse:
            return self.get(
                path="/projects",
                query_params={
                    "page": str(page),
                    "limit": str(LIST_PAGINATION_LIMIT),
                },
                filter=filter,
                order_by=order_by,
            )

        return self.build_iterator(fetcher=fetcher, limit=limit, resource=ListedProject)

    def update(
        self,
        *,
        id: str,
        name: Optional[str] = NOT_SET,
        columns: Optional[List[Dict[str, Any]]] = NOT_SET,
        tags: Optional[List[str]] = NOT_SET,
    ) -> "ProjectDetail":
        """Updates a project you have previously created.

        :param id: The project identifier.
        :param name: Project name, as displayed in the user interface.
        :param columns: Columns of a project.
        :param tags: Tags assigned to this project.
        :raises caplena.api.ApiException: An API exception.
        """
        json = self.api.build_payload(
            name=name,
            tags=tags,
            columns=columns,
        )

        response = self.patch(path="/projects/{id}", path_params={"id": id}, json=json)
        return self.build_response(response, resource=ProjectDetail)

    def append_rows(
        self,
        *,
        id: str,
        rows: List[Dict[str, Any]],
    ) -> "RowsAppend":
        """Appends multiple rows to a previously created project. It is possible to append a
        maximum of 20 rows in a single request.

        :param id: The project identifier.
        :param rows: The rows to append to the specified project.
        :raises caplena.api.ApiException: An API exception.
        """
        response = self.post(
            path="/projects/{id}/rows/bulk",
            path_params={"id": id},
            json=rows,
            allowed_codes={202},
        )

        return self.build_response(response, resource=RowsAppend)

    def append_row(
        self,
        *,
        id: str,
        columns: List[Dict[str, Any]],
    ) -> "Row":
        """Appends a single row to a previously created project.

        :param id: The project identifier.
        :param columns: The columns for the new row.
        :raises caplena.api.ApiException: An API exception.
        """
        json = self.api.build_payload(columns=columns)
        response = self.post(
            path="/projects/{id}/rows",
            path_params={"id": id},
            json=json,
        )

        return self.build_response(response, resource=Row, metadata={"project": id})

    def list_rows(
        self,
        *,
        id: str,
        limit: Optional[int] = None,
        filter: Optional[RowsFilter] = None,
    ) -> "CaplenaIterator[Row]":
        """Returns a list of all rows you have previously created for this project. The rows are returned in
        sorted order, with the least recently added row appearing first.

        :param id: The project identifier.
        :param limit: Number of results returned per page. If unspecified, will return all results.
        :param filter: Filters to apply to this request. If omitted, no filters are applied.
        :raises caplena.api.ApiException: An API exception.
        """

        def fetcher(page: int) -> HttpResponse:
            return self.get(
                path="/projects/{id}/rows",
                path_params={"id": id},
                query_params={
                    "page": str(page),
                    "limit": str(LIST_PAGINATION_LIMIT),
                },
                filter=filter,
            )

        return self.build_iterator(
            fetcher=fetcher, limit=limit, resource=Row, metadata={"project": id}
        )

    def retrieve_row(self, *, p_id: str, r_id: str) -> "Row":
        """Retrieves a row for a project you have previously created.

        :param p_id: The project identifier.
        :param r_id: The row identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        response = self.get(
            path="/projects/{p_id}/rows/{r_id}",
            path_params={"p_id": p_id, "r_id": r_id},
        )
        return self.build_response(response, resource=Row, metadata={"project": id})

    def remove_row(self, *, p_id: str, r_id: str) -> None:
        """Removes a previously created row.

        :param p_id: The project identifier.
        :param r_id: The row identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        self.delete(path="/projects/{p_id}/rows/{r_id}", path_params={"p_id": p_id, "r_id": r_id})

    def update_row(
        self,
        *,
        p_id: str,
        r_id: str,
        columns: List[Dict[str, Any]],
    ) -> "Row":
        """Updates a row you have previously created.

        :param p_id: The project identifier.
        :param r_id: The row identifier.
        :param columns: Columns for this row.
        :raises caplena.api.ApiException: An API exception.
        """
        json = self.api.build_payload(columns=columns)

        response = self.patch(
            path="/projects/{p_id}/rows/{r_id}", path_params={"p_id": p_id, "r_id": r_id}, json=json
        )
        return self.build_response(response, resource=Row, metadata={"project": id})


# --- Resources & Objects--- #


class ProjectDetail(BaseResource[ProjectsController]):
    """The project detail resource."""

    class Column(BaseObject[ProjectsController]):
        __fields__ = {"ref", "name", "type"}
        __mutable__ = {"name"}

        ref: str
        """Human-readable identifier for this column. The reference field
        is immutable and is unique among all columns within the same project.
        """

        name: str
        """Human-readable name for this column."""

        type: Literal["numerical", "boolean", "text", "date", "any", "text_to_analyze"]
        """Type of this column."""

        def modified_dict(self) -> Any:
            modified: Any = super().modified_dict()
            if modified != NOT_SET:
                modified["ref"] = self.ref
                modified["type"] = self.type

            return modified

        def __repr__(self) -> str:
            return f"ProjectColumn(ref={self.ref}, type={self.type}, name={self.name})"

    class TextToAnalyze(Column):
        class Topic(BaseResource[ProjectsController]):
            class Sentiment(BaseObject[ProjectsController]):
                __fields__ = {"code", "label"}

                code: int
                """Code for this topic sentiment."""

                label: str
                """Label for this topic sentiment."""

                def __repr__(self) -> str:
                    return f"TopicSentiment(code={self.code}, label={self.label})"

            __fields__ = {
                "label",
                "category",
                "color",
                "description",
                "sentiment_enabled",
                "sentiment_neutral",
                "sentiment_negative",
                "sentiment_positive",
            }

            label: str
            """Label for this topic."""

            category: str
            """Category for this topic"""

            color: str
            """Color for this topic."""

            description: str
            """Description for this topic."""

            sentiment_enabled: bool
            """If enabled, Caplena will assign the topic with a corresponding sentiment."""

            sentiment_neutral: Sentiment
            """Neutral topic sentiment."""

            sentiment_negative: Sentiment
            """Negative topic sentiment."""

            sentiment_positive: Sentiment
            """Positive topic sentiment."""

            def __repr__(self) -> str:
                return (
                    f"{self.__class__.__name__}(label={self.label}, category={self.category}, "
                    f"sentiment_enabled={self.sentiment_enabled})"
                )

            @classmethod
            def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail.TextToAnalyze.Topic":
                obj["sentiment_neutral"] = cls.Sentiment.parse_obj(obj["sentiment_neutral"])
                obj["sentiment_negative"] = cls.Sentiment.parse_obj(obj["sentiment_negative"])
                obj["sentiment_positive"] = cls.Sentiment.parse_obj(obj["sentiment_positive"])

                return super().parse_obj(obj)

        class Metadata(BaseObject[ProjectsController]):
            class LearnsForm(BaseObject[ProjectsController]):
                __fields__ = {"project", "ref"}
                __mutable__ = {"project", "ref"}

                project: str
                """Base project that this column learns from."""

                ref: str
                """Column identifier that this column learns from."""

            __fields__ = {
                "reviewed_count",
                "learns_from",
            }
            __mutable__ = {"learns_from"}

            reviewed_count: int
            """Number of reviewed rows for this column."""

            learns_from: Optional[LearnsForm]
            """Base column that this column learns from."""

        __fields__ = {"ref", "name", "type", "description", "topics", "metadata"}
        __mutable__ = {"name", "description"}

        type: Literal["text_to_analyze"]
        """Type of this column."""

        description: Optional[str] = ""
        """Column description displayed for this column."""

        topics: CaplenaList[Topic]
        """List of topics associated with this column."""

        metadata: Metadata
        """Metadata associated with this column."""

        def __repr__(self) -> str:
            return f"ProjectColumn(ref={self.ref}, type={self.type}, name={self.name}, topics={self.topics.__repr__()})"

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail.TextToAnalyze":
            obj["topics"] = CaplenaList(
                values=[cls.Topic.parse_obj(topic) for topic in obj["topics"]]
            )
            obj["metadata"] = cls.Metadata.parse_obj(obj["metadata"])
            return super().parse_obj(obj)

    class Auxiliary(Column):

        type: Literal["numerical", "boolean", "text", "date", "any"]
        """Type of this column."""

    __fields__ = {
        "name",
        "owner",
        "tags",
        "upload_status",
        "language",
        "columns",
        "created",
        "last_modified",
        "translation_status",
        "translation_engine",
    }
    __mutable__ = {"name", "tags"}

    name: str
    """Name of this project."""

    owner: str
    """Identifier of the user that owns this project."""

    tags: List[str]
    """Tags associated with this project."""

    upload_status: Literal["pending", "in_progress", "succeeded", "failed"]
    """Current upload status of this project."""

    language: str
    """Base language for this project as an ISO-639-1 Code"""

    columns: CaplenaList[Column]
    """Columns for this projects."""

    created: datetime
    """Timestamp at which the project was created."""

    last_modified: datetime
    """Timestamp at which the project was last updated."""

    translation_status: Optional[str]
    """Current translation status for this project."""

    translation_engine: Optional[str]
    """Translation engine used for translating :code:`text_to_analyze` columns."""

    anonymize_pii: Optional[Dict[str, List[str]]]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name}, columns={self.columns.__repr__()})"

    def remove(self) -> None:
        """Removes this project.

        :raises caplena.api.ApiException: An API exception.
        """
        self.controller.remove(id=self.id)

    def append_row(self, *, columns: List[Dict[str, Any]]) -> "Row":
        """Appends a single row to this project.

        :param columns: The columns for the new row.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.append_row(id=self.id, columns=columns)

    def append_rows(self, *, rows: List[Dict[str, Any]]) -> "RowsAppend":
        """Appends multiple rows to this project. It is possible to append a
        maximum of 20 rows in a single request.

        :param rows: The rows to append to the specified project.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.append_rows(id=self.id, rows=rows)

    def list_rows(
        self,
        *,
        limit: Optional[int] = None,
        filter: Optional[RowsFilter] = None,
    ) -> "CaplenaIterator[Row]":
        """Returns a list of all rows you have previously created for this project. The rows are returned in
        sorted order, with the least recently added row appearing first.

        :param limit: Number of results returned per page.
        :param filter: Filters to apply to this request. If omitted, no filters are applied.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.list_rows(id=self.id, limit=limit, filter=filter)

    def retrieve_row(self, *, id: str) -> "Row":
        """Retrieves a previously created row for this project.

        :param id: The row identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.retrieve_row(p_id=self.id, r_id=id)

    def refresh(self) -> None:
        """Refreshes the properties of this project.

        :raises caplena.api.ApiException: An API exception.
        """
        project = self.controller.retrieve(id=self.id)
        self._refresh_from(attrs=project._attrs)

    def save(self) -> None:
        """Saves the unpersisted properties of this project.

        :raises caplena.api.ApiException: An API exception.
        """
        modified_dict = self.modified_dict()
        if modified_dict != NOT_SET:
            project = self.controller.update(id=self.id, **modified_dict)
            self._refresh_from(attrs=project._attrs)

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "ProjectDetail":
        obj["columns"] = CaplenaList(
            values=[
                cls.TextToAnalyze.parse_obj(column)
                if column["type"] == "text_to_analyze"
                else cls.Auxiliary.parse_obj(column)
                for column in obj["columns"]
            ]
        )
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)


class ListedProject(BaseResource[ProjectsController]):
    """The project list resource."""

    __fields__ = {
        "name",
        "owner",
        "tags",
        "upload_status",
        "language",
        "created",
        "last_modified",
        "translation_status",
        "translation_engine",
    }
    __mutable__ = {"name", "tags"}

    name: str
    """Name of this project."""

    owner: str
    """Identifier of the user that owns this project."""

    tags: List[str]
    """Tags associated with this project."""

    upload_status: Literal["pending", "in_progress", "succeeded", "failed"]
    """Current upload status of this project."""

    language: str
    """Base language for this project as an ISO-639-1 Code"""

    created: datetime
    """Timestamp at which the project was created."""

    last_modified: datetime
    """Timestamp at which the project was last updated."""

    translation_status: Optional[str]
    """Current translation status for this project."""

    translation_engine: Optional[str]
    """Translation engine used for translating :code:`text_to_analyze` columns."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name})"

    def remove(self) -> None:
        """Removes this project.

        :raises caplena.api.ApiException: An API exception.
        """
        self.controller.remove(id=self.id)

    def append_row(self, *, columns: List[Dict[str, Any]]) -> "Row":
        """Appends a single row to this project.

        :param columns: The columns for the new row.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.append_row(id=self.id, columns=columns)

    def append_rows(self, *, rows: List[Dict[str, Any]]) -> "RowsAppend":
        """Appends multiple rows to this project. It is possible to append a
        maximum of 20 rows in a single request.

        :param rows: The rows to append to the specified project.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.append_rows(id=self.id, rows=rows)

    def list_rows(
        self,
        *,
        limit: Optional[int] = None,
        filter: Optional[RowsFilter] = None,
    ) -> "CaplenaIterator[Row]":
        """Returns a list of all rows you have previously created for this project. The rows are returned in
        sorted order, with the least recently added row appearing first.

        :param limit: Number of results returned per page.
        :param filter: Filters to apply to this request. If omitted, no filters are applied.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.list_rows(id=self.id, limit=limit, filter=filter)

    def retrieve_row(self, *, id: str) -> "Row":
        """Retrieves a previously created row for this project.

        :param id: The row identifier.
        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.retrieve_row(p_id=self.id, r_id=id)

    def refresh(self) -> None:
        """Refreshes the properties of this project.

        :raises caplena.api.ApiException: An API exception.
        """
        project = self.controller.retrieve(id=self.id)
        self._refresh_from(attrs=project._attrs)

    def save(self) -> None:
        """Saves the unpersisted properties of this project.

        :raises caplena.api.ApiException: An API exception.
        """
        modified_dict = self.modified_dict()
        if modified_dict != NOT_SET:
            project = self.controller.update(id=self.id, **modified_dict)
            self._refresh_from(attrs=project._attrs)

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "ListedProject":
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)


class RowsAppend(BaseObject[ProjectsController]):
    """The bulk row create response object."""

    __fields__ = {"status", "queued_rows_count", "estimated_minutes"}

    status: Literal["pending"]
    """Status of the bulk append operation."""

    queued_rows_count: int
    """Number of rows that were queued for appending."""

    estimated_minutes: float
    """Estimation in minutes about how long this bulk opertion will approximately take."""


class Row(BaseResource[ProjectsController]):
    """The Row resource."""

    class Column(BaseObject[ProjectsController]):
        __fields__ = {"ref", "type", "value"}
        __mutable__ = {"value"}

        ref: str
        """Human-readable identifier for this column."""

        type: Literal["numerical", "boolean", "text", "date", "any", "text_to_analyze"]
        """Type of this column."""

        value: Union[int, float, bool, None, str, datetime]
        """Value assigned to this column."""

        def __repr__(self) -> str:
            return f"RowColumn(ref={self.ref}, type={self.type}, value={self.value})"

        def modified_dict(self) -> Any:
            modified: Any = super().modified_dict()
            if modified != NOT_SET:
                modified["ref"] = self.ref

            return modified

    class NumericalColumn(Column):
        type: Literal["numerical"]
        """Type of this column."""

        value: Optional[Union[int, float]]
        """Numerical value assigned to this column."""

    class BooleanColumn(Column):
        type: Literal["boolean"]
        """Type of this column."""

        value: Optional[bool]
        """Boolean value assigned to this column."""

    class DateColumn(Column):
        type: Literal["date"]
        """Type of this column."""

        value: Optional[datetime]
        """ISO 8601 datetime or date value assigned to this column."""

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "Row.DateColumn":
            if obj["value"] is not None:
                obj["value"] = Helpers.from_rfc3339_datetime(obj["value"])

            return super().parse_obj(obj)

    class AnyColumn(Column):
        type: Literal["any"]
        """Type of this column."""

        value: None
        """Any value assigned to this column."""

    class TextColumn(Column):
        type: Literal["text"]
        """Type of this column."""

        value: str
        """Text value assigned to this column."""

    class TextToAnalyzeColumn(Column):
        class Topic(BaseObject[ProjectsController]):
            __fields__ = {"id", "label", "category", "code", "sentiment_label", "sentiment"}
            __mutable__ = {"sentiment"}

            id: str
            """Unique identifier for this topic."""

            label: str
            """Label for this topic."""

            category: str
            """Category for this topic."""

            code: int
            """Code for the inferred topic sentiment. If sentiment is disabled, the neutral
            topic sentiment :code:`code` will be used."""

            sentiment_label: str
            """Label for the inferred topic sentiment. If sentiment is disabled, the neutral
            topic sentiment :code:`label` will be used."""

            sentiment: Literal["neutral", "positive", "negative", "any"]
            """Inferred sentiment for this column value."""

            def __repr__(self) -> str:
                return f"{self.__class__.__name__}(label={self.label}, category={self.category}, sentiment={self.sentiment})"

        __fields__ = {
            "ref",
            "type",
            "value",
            "was_reviewed",
            "sentiment_overall",
            "source_language",
            "translated_value",
            "topics",
        }
        __mutable__ = {"value", "was_reviewed", "topics"}

        type: Literal["text_to_analyze"]
        """Type of this column."""

        value: str
        """Text value assigned to this column."""

        was_reviewed: Optional[bool]
        """Indicates whether the row value for this column has been reviewed."""

        sentiment_overall: Optional[Literal["neutral", "positive", "negative"]]
        """Inferred overall sentiment for this column."""

        source_language: Optional[str]
        """Source language of this value as an ISO-639-1 Code (https://cloud.google.com/translate/docs/languages)."""

        translated_value: Optional[str]
        """Translated value if translation is enabled for this column."""

        topics: CaplenaList[Topic]
        """Topics matching the value of this column. If no topics match, an empty array is returned."""

        def __repr__(self) -> str:
            return (
                f"RowColumn(ref={self.ref}, type={self.type}, value={self.value[:50]}..., "
                f"sentiment_overall={self.sentiment_overall}, topics={self.topics.__repr__()})"
            )

        @classmethod
        def parse_obj(cls, obj: Dict[str, Any]) -> "Row.TextToAnalyzeColumn":
            obj["topics"] = CaplenaList(
                values=[cls.Topic.parse_obj(topic) for topic in obj["topics"]]
            )
            return super().parse_obj(obj)

    __fields__ = {"created", "last_modified", "columns"}

    created: datetime
    """Timestamp at which this row was created."""

    last_modified: datetime
    """Timestamp at which this row was last updated."""

    columns: CaplenaList[Column]
    """Columns for this row."""

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, columns={self.columns.__repr__()})"

    def remove(self) -> None:
        """Removes this row.

        :raises caplena.api.ApiException: An API exception.
        """
        self.controller.remove_row(p_id=self._metadata["project"], r_id=self.id)

    def refresh(self) -> None:
        """Refreshes the properties of this row.

        :raises caplena.api.ApiException: An API exception.
        """
        row = self.controller.retrieve_row(p_id=self._metadata["project"], r_id=self.id)
        self._refresh_from(attrs=row._attrs)

    def save(self) -> None:
        """Saves the unpersisted properties of this row.

        :raises caplena.api.ApiException: An API exception.
        """
        modified_dict = self.modified_dict()
        if modified_dict != NOT_SET:
            row = self.controller.update_row(
                p_id=self._metadata["project"],
                r_id=self.id,
                **modified_dict,
            )
            self._refresh_from(attrs=row._attrs)

    def retrieve_project(self) -> "ProjectDetail":
        """Retrieves the project that this row belongs to.

        :raises caplena.api.ApiException: An API exception.
        """
        return self.controller.retrieve(id=self._metadata["project"])

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "Row":
        type_to_column = {
            "numerical": cls.NumericalColumn,
            "boolean": cls.BooleanColumn,
            "date": cls.DateColumn,
            "any": cls.AnyColumn,
            "text": cls.TextColumn,
            "text_to_analyze": cls.TextToAnalyzeColumn,
        }
        obj["columns"] = CaplenaList(
            values=[type_to_column[column["type"]].parse_obj(column) for column in obj["columns"]]
        )
        obj["created"] = Helpers.from_rfc3339_datetime(obj["created"])
        obj["last_modified"] = Helpers.from_rfc3339_datetime(obj["last_modified"])

        return super().parse_obj(obj)



from datetime import datetime

from typing_extensions import Literal

from caplena.api import ApiFilter, ZeroOrMany

DateRange = Literal[
    "all_time", "this_month", "last_month", "this_quarter", "last_quarter", "this_year", "last_year"
]


class ProjectsFilter(ApiFilter):
    """The filter that can be used to filter projects.

    :param constraints: The internal filter constraints. Should never be manually given.
    :param has_conjunction: The internal conjunction boolean. Should never be manually given.
    """

    @classmethod
    def name(
        cls,
        *,
        exact__i: ZeroOrMany[str] = None,
        contains__i: ZeroOrMany[str] = None,
    ) -> "ProjectsFilter":
        """Allows filtering results based on text fields. For example, filtering with
        :code:`ProjectsFilter.name(contains__i='reviews')` only returns results containing the
        text reviews for the specified field.

        :param exact__i: Allows filtering on exact, case-insensitive text.
        :param contains__i: Allows filtering on partial, case-insensitive text.
        """
        return cls.construct(
            name="name",
            filters={
                "exact.i": exact__i,
                "contains.i": contains__i,
            },
        )

    @classmethod
    def owner(
        cls,
        *,
        id: ZeroOrMany[str],
        email__exact__i: ZeroOrMany[str],
        email__contains__i: ZeroOrMany[str],
    ) -> "ProjectsFilter":
        """Allows filtering projects based on their owner. For example, filtering with
        :code:`ProjectsFilter.owner(email__exact__i='katy@acme.com')` only returns
        projects that belong to :code:`katy@acme.com`.

        :param id: Allows filtering on project owner identifiers.
        :param email__exact__i: Allows filtering on exact, case-insensitive email addresses of project owners.
        :param email__contains__i: Allows filtering on partial, case-insensitive email addresses of project owners.
        """
        return cls.construct(
            name="owner",
            filters={
                "id": id,
                "email.exact.i": email__exact__i,
                "email.contains.i": email__contains__i,
            },
        )

    @classmethod
    def tags(cls, tag: ZeroOrMany[str]) -> "ProjectsFilter":
        """Allows filtering projects based on the tags they contain. For example, filtering with
        :code:`ProjectsFilter.tags('tag1') & ProjectsFilter.tags('tag2')` returns all projects
        that are tagged with both :code:`tag1` and :code:`tag2`.

        :param tag: Allows filtering on project tags.
        """
        return cls.construct(
            name="tags",
            filters={
                cls.DEFAULT: tag,
            },
        )

    @classmethod
    def upload_status(cls, upload_status: ZeroOrMany[str]) -> "ProjectsFilter":
        """Allows filtering projects based on their upload status. For example, filtering
        with :code:`ProjectsFilter.upload_status(['pending', 'failed'])` returns all
        projects that are pending or have failed.

        :param upload_status: Allows filtering on upload status.
        """
        return cls.construct(
            name="upload_status",
            filters={
                cls.DEFAULT: upload_status,
            },
        )

    @classmethod
    def language(cls, language: ZeroOrMany[str]) -> "ProjectsFilter":
        """Allows filtering projects based on their language. For example, filtering with
        :code:`ProjectsFilter.language(['en', 'es'])` returns all English and Spanish projects.

        :param language: Allows filtering on project language.
        """
        return cls.construct(
            name="language",
            filters={
                cls.DEFAULT: language,
            },
        )

    @classmethod
    def translation_status(cls, translation_status: ZeroOrMany[str]) -> "ProjectsFilter":
        """Allows filtering projects based on their translation status. For example, filtering
        with :code:`ProjectsFilter.translation_status(['disabled', 'pending'])` returns all
        projects that have translation disabled or translation is still pending.

        :param translation_status: Allows filtering on translation status.
        """
        return cls.construct(
            name="translation_status",
            filters={
                cls.DEFAULT: translation_status,
            },
        )

    @classmethod
    def translation_engine(cls, translation_engine: ZeroOrMany[str]) -> "ProjectsFilter":
        """Allows filtering projects based on their translation engine. For example, filtering
        with :code:`ProjectsFilter.translation_engine('google_translate')` returns all projects
        that have Google Translate enabled.

        :param translation_engine: Allows filtering on translation engine.
        """
        return cls.construct(
            name="translation_engine",
            filters={
                cls.DEFAULT: translation_engine,
            },
        )

    @classmethod
    def created(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        range: ZeroOrMany[DateRange] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "ProjectsFilter":
        """Allows filtering results based on date fields. For example, filtering with
        :code:`ProjectsFilter.created(year=2022, month=1)` returns all results for
        January 2022 for the specified field.

        :param gte: Greater than or equal to filter for date times.
        :param gt: Greater than filter for date times.
        :param lte: Less than or equal to filter for date times.
        :param lt: Less than filter for date times.
        :param range: Range filter for date times.
        :param year: Allows filtering on the year.
        :param year__gte: Greater than or equal to filter on the year.
        :param year__gt: Greater than filter on the year.
        :param year__lte: Less than or equal to filter on the year.
        :param year__lt: Less than filter on the year.
        :param month: Allows filtering on the month of the year.
        :param month__gte: Greater than or equal to filter on the month of the year.
        :param month__gt: Greater than filter on the month of the year.
        :param month__lte: Less than or equal to filter on the month of the year.
        :param month__lt: Greater than filter on the month of the year.
        :param day: Allows filtering on the day of the month.
        :param day__gte: Greater than or equal to filter on the day of the month.
        :param day__gt: Greater than filter on the day of the month.
        :param day__lte: Less than or equal to filter on the day of the month.
        :param day__lt: Greater than filter on the day of the month.
        """
        return cls.construct(
            name="created",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "range": range,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )

    @classmethod
    def last_modified(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        range: ZeroOrMany[DateRange] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "ProjectsFilter":
        """Allows filtering results based on date fields. For example, filtering with
        :code:`ProjectsFilter.last_modified(year=2022, month=1)` returns all results for
        January 2022 for the specified field.

        :param gte: Greater than or equal to filter for date times.
        :param gt: Greater than filter for date times.
        :param lte: Less than or equal to filter for date times.
        :param lt: Less than filter for date times.
        :param range: Range filter for date times.
        :param year: Allows filtering on the year.
        :param year__gte: Greater than or equal to filter on the year.
        :param year__gt: Greater than filter on the year.
        :param year__lte: Less than or equal to filter on the year.
        :param year__lt: Less than filter on the year.
        :param month: Allows filtering on the month of the year.
        :param month__gte: Greater than or equal to filter on the month of the year.
        :param month__gt: Greater than filter on the month of the year.
        :param month__lte: Less than or equal to filter on the month of the year.
        :param month__lt: Greater than filter on the month of the year.
        :param day: Allows filtering on the day of the month.
        :param day__gte: Greater than or equal to filter on the day of the month.
        :param day__gt: Greater than filter on the day of the month.
        :param day__lte: Less than or equal to filter on the day of the month.
        :param day__lt: Greater than filter on the day of the month.
        """
        return cls.construct(
            name="last_modified",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "range": range,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )


class RowsFilter(ApiFilter):
    """The filter that can be used to filter rows.

    :param constraints: The internal filter constraints. Should never be manually given.
    :param has_conjunction: The internal conjunction boolean. Should never be manually given.
    """

    class Columns:
        """Allows filtering rows based on the values of its columns."""

        @staticmethod
        def numerical(
            *,
            ref: str,
            exact: ZeroOrMany[float] = None,
            gte: ZeroOrMany[float] = None,
            gt: ZeroOrMany[float] = None,
            lte: ZeroOrMany[float] = None,
            lt: ZeroOrMany[float] = None,
        ) -> "RowsFilter":
            """Allows filtering on numerical columns.

            :param ref: The numerical column reference.
            :param exact: Exact filter for numerical values.
            :param gte: Greater than or equal to filter for numerical values.
            :param gt: Greater than filter for numerical values.
            :param lte: Less than or equal to filter for numerical values.
            :param lt: Less than filter for numerical values.
            """
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[numerical]": exact,
                    f"{ref}[numerical].gte": gte,
                    f"{ref}[numerical].gt": gt,
                    f"{ref}[numerical].lte": lte,
                    f"{ref}[numerical].lt": lt,
                },
            )

        @classmethod
        def boolean(
            cls,
            *,
            ref: str,
            exact: ZeroOrMany[bool] = None,
        ) -> "RowsFilter":
            """Allows filtering on boolean columns.

            :param ref: The boolean column reference.
            :param exact: Exact filter for boolean values.
            """
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[boolean]": exact,
                },
            )

        @classmethod
        def text(
            cls,
            *,
            ref: str,
            exact__i: ZeroOrMany[str] = None,
            contains__i: ZeroOrMany[str] = None,
        ) -> "RowsFilter":
            """Allows filtering on text columns.

            :param ref: The text column reference.
            :param exact__i: Allows filtering on exact, case-insensitive text values.
            :param contains__i: Allows filtering on partial, case-insensitive text values.
            """
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[text].exact.i": exact__i,
                    f"{ref}[text].contains.i": contains__i,
                },
            )

        @classmethod
        def date(
            cls,
            *,
            ref: str,
            gte: ZeroOrMany[datetime] = None,
            gt: ZeroOrMany[datetime] = None,
            lte: ZeroOrMany[datetime] = None,
            lt: ZeroOrMany[datetime] = None,
            range: ZeroOrMany[DateRange] = None,
        ) -> "RowsFilter":
            """Allows filtering on date columns.

            :param gte: Greater than or equal to filter for date times.
            :param gt: Greater than filter for date times.
            :param lte: Less than or equal to filter for date times.
            :param lt: Less than filter for date times.
            :param range: Range filter for date times.
            """
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[date].gte": gte,
                    f"{ref}[date].gt": gt,
                    f"{ref}[date].lte": lte,
                    f"{ref}[date].lt": lt,
                    f"{ref}[date].range": range,
                },
            )

        @classmethod
        def text_to_analyze(
            cls,
            *,
            ref: str,
            exact__i: ZeroOrMany[str] = None,
            contains__i: ZeroOrMany[str] = None,
            was_reviewed: ZeroOrMany[bool] = None,
            source_language: ZeroOrMany[str] = None,
            translated_value__exact__i: ZeroOrMany[str] = None,
            translated_value__contains__i: ZeroOrMany[str] = None,
        ) -> "RowsFilter":
            """Allows filtering on verbatim columns.

            :param ref: The verbatim column reference.
            :param exact__i: Allows filtering on exact, case-insensitive verbatims.
            :param contains__i: Allows filtering on partial, case-insensitive verbatims.
            :param was_reviewed: Allows filtering on reviewed columns.
            :param source_language: Allows filtering on the source language of this column.
            :param translated_value__exact__i: Allows filtering on exact, case-insensitive translated text verbatims.
            :param translated_value__contains__i: Allows filtering on partial, case-insensitive translated text verbatims.
            """
            return RowsFilter.construct(
                name="columns",
                filters={
                    f"{ref}[text_to_analyze].exact.i": exact__i,
                    f"{ref}[text_to_analyze].contains.i": contains__i,
                    f"{ref}[text_to_analyze].was_reviewed": was_reviewed,
                    f"{ref}[text_to_analyze].source_language": source_language,
                    f"{ref}[text_to_analyze].translated_value.exact.i": translated_value__exact__i,
                    f"{ref}[text_to_analyze].translated_value.contains.i": translated_value__contains__i,
                },
            )

    @classmethod
    def created(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        range: ZeroOrMany[DateRange] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "RowsFilter":
        """Allows filtering results based on date fields. For example, filtering with
        :code:`RowsFilter.created(year=2022, month=1)` returns all results for
        January 2022 for the specified field.

        :param gte: Greater than or equal to filter for date times.
        :param gt: Greater than filter for date times.
        :param lte: Less than or equal to filter for date times.
        :param lt: Less than filter for date times.
        :param range: Range filter for date times.
        :param year: Allows filtering on the year.
        :param year__gte: Greater than or equal to filter on the year.
        :param year__gt: Greater than filter on the year.
        :param year__lte: Less than or equal to filter on the year.
        :param year__lt: Less than filter on the year.
        :param month: Allows filtering on the month of the year.
        :param month__gte: Greater than or equal to filter on the month of the year.
        :param month__gt: Greater than filter on the month of the year.
        :param month__lte: Less than or equal to filter on the month of the year.
        :param month__lt: Greater than filter on the month of the year.
        :param day: Allows filtering on the day of the month.
        :param day__gte: Greater than or equal to filter on the day of the month.
        :param day__gt: Greater than filter on the day of the month.
        :param day__lte: Less than or equal to filter on the day of the month.
        :param day__lt: Greater than filter on the day of the month.
        """
        return cls.construct(
            name="created",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "range": range,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )

    @classmethod
    def last_modified(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        gt: ZeroOrMany[datetime] = None,
        lte: ZeroOrMany[datetime] = None,
        lt: ZeroOrMany[datetime] = None,
        range: ZeroOrMany[DateRange] = None,
        year: ZeroOrMany[int] = None,
        year__gte: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lte: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        month__gte: ZeroOrMany[int] = None,
        month__gt: ZeroOrMany[int] = None,
        month__lte: ZeroOrMany[int] = None,
        month__lt: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
        day__gte: ZeroOrMany[int] = None,
        day__gt: ZeroOrMany[int] = None,
        day__lte: ZeroOrMany[int] = None,
        day__lt: ZeroOrMany[int] = None,
    ) -> "RowsFilter":
        """Allows filtering results based on date fields. For example, filtering with
        :code:`RowsFilter.last_modified(year=2022, month=1)` returns all results for
        January 2022 for the specified field.

        :param gte: Greater than or equal to filter for date times.
        :param gt: Greater than filter for date times.
        :param lte: Less than or equal to filter for date times.
        :param lt: Less than filter for date times.
        :param range: Range filter for date times.
        :param year: Allows filtering on the year.
        :param year__gte: Greater than or equal to filter on the year.
        :param year__gt: Greater than filter on the year.
        :param year__lte: Less than or equal to filter on the year.
        :param year__lt: Less than filter on the year.
        :param month: Allows filtering on the month of the year.
        :param month__gte: Greater than or equal to filter on the month of the year.
        :param month__gt: Greater than filter on the month of the year.
        :param month__lte: Less than or equal to filter on the month of the year.
        :param month__lt: Greater than filter on the month of the year.
        :param day: Allows filtering on the day of the month.
        :param day__gte: Greater than or equal to filter on the day of the month.
        :param day__gt: Greater than filter on the day of the month.
        :param day__lte: Less than or equal to filter on the day of the month.
        :param day__lt: Greater than filter on the day of the month.
        """
        return cls.construct(
            name="last_modified",
            filters={
                "gte": gte,
                "gt": gt,
                "lte": lte,
                "lt": lt,
                "range": range,
                "year": year,
                "year.gte": year__gte,
                "year.gt": year__gt,
                "year.lte": year__lte,
                "year.lt": year__lt,
                "month": month,
                "month.gte": month__gte,
                "month.gt": month__gt,
                "month.lte": month__lte,
                "month.lt": month__lt,
                "day": day,
                "day.gte": day__gte,
                "day.gt": day__gt,
                "day.lte": day__lte,
                "day.lt": day__lt,
            },
        )



from typing import Iterable, Type, Union

from caplena.api import ApiBaseUri, ApiVersion
from caplena.configuration import Configuration
from caplena.controllers import ProjectsController
from caplena.http.http_client import HttpClient, HttpMethod, HttpRetry
from caplena.http.requests_http_client import RequestsHttpClient
from caplena.logging.logger import LoggingLevel


class Client:
    """Represents a client connection that connects to Caplena. This class is used
    to interact with the Caplena REST API.

    :param api_key: The API key to use for making requests.
    :param api_base_uri: The API Base URI to use, defaults to :code:`https://api.caplena.com/v2`.
    :type api_base_uri: ApiBaseUri
    :param api_version: The API Version to use, defaults to :code:`2022-06-13`.
    :type api_version: ApiVersion
    :param timeout: The maximum number of seconds before the request times out, defaults to :code:`120`.
    :param max_retries: The maximum number of times the request is retried before giving up, defaults to :code:`0`.
    :param backoff_factor: The backoff factor to apply between attempts, defaults to :code:`2`.
    :param retry_status_codes: A set of HTTP status codes that we should retry on, defaults to
        :code:`{408, 409, 413, 429, 500, 502, 503, 504}`.
    :param retry_methods: A set of HTTP methods that we should retry on, defaults to :code:`{GET, PUT, HEAD}`.
    :type retry_methods: Iterable[HttpMethod]
    :param http_client: The HTTP client class or instance to use for making requests, defaults to :code:`RequestsHttpClient`.
        If an HTTP class is given, the factory method :code:`build_http_client` is used to create an instance.
    :type http_client: Union[HttpClient, Type[HttpClient]]
    :param logging_level: The level of events to log out to console, defaults to :code:`WARNING`.
    :type logging_level: LoggingLevel
    """

    @property
    def projects(self) -> ProjectsController:
        """The projects controller, encapsulating all project actions."""
        return self._projects_controller

    @property
    def config(self) -> Configuration:
        return self._config

    def __init__(
        self,
        api_key: str,
        *,
        api_base_uri: ApiBaseUri = ApiBaseUri.PRODUCTION,
        api_version: ApiVersion = ApiVersion.VER_2022_06_20,
        timeout: int = HttpClient.DEFAULT_TIMEOUT,
        max_retries: int = HttpRetry.DEFAULT_MAX_RETRIES,
        backoff_factor: float = HttpRetry.DEFAULT_BACKOFF_FACTOR,
        retry_status_codes: Iterable[int] = HttpRetry.DEFAULT_STATUS_CODES_TO_RETRY,
        retry_methods: Iterable[HttpMethod] = HttpRetry.DEFAULT_ALLOWED_METHOD,
        http_client: Union[Type[HttpClient], HttpClient] = RequestsHttpClient,
        logging_level: LoggingLevel = LoggingLevel.WARNING,
    ):
        self._config = Configuration(
            api_key=api_key,
            http_client=http_client,
            api_base_uri=api_base_uri,
            api_version=api_version,
            timeout=timeout,
            max_retries=max_retries,
            backoff_factor=backoff_factor,
            retry_status_codes=retry_status_codes,
            retry_methods=retry_methods,
            logging_level=logging_level,
        )

        self._projects_controller = ProjectsController(config=self._config)



import unittest
from datetime import datetime, timezone
from typing import Any, ClassVar, List, Tuple

from caplena.api import ApiBaseUri, ApiFilter, ApiRequestor, ApiVersion, ZeroOrMany
from caplena.http.http_client import HttpClient
from caplena.http.requests_http_client import RequestsHttpClient
from caplena.logging.default_logger import DefaultLogger
from caplena.logging.logger import Logger


class Pf(ApiFilter):
    @classmethod
    def created(
        cls,
        *,
        year: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
    ) -> "Pf":
        return cls.construct(
            name="created",
            filters={
                "year": year,
                "year.gt": year__gt,
                "year.lt": year__lt,
                "month": month,
                "day": day,
            },
        )

    @classmethod
    def last_modified(
        cls,
        *,
        gte: ZeroOrMany[datetime] = None,
        year: ZeroOrMany[int] = None,
        year__gt: ZeroOrMany[int] = None,
        year__lt: ZeroOrMany[int] = None,
        month: ZeroOrMany[int] = None,
        day: ZeroOrMany[int] = None,
    ) -> "Pf":
        return cls.construct(
            name="last_modified",
            filters={
                "gte": gte,
                "year": year,
                "year.gt": year__gt,
                "year.lt": year__lt,
                "month": month,
                "day": day,
            },
        )

    @classmethod
    def tags(cls, tag: ZeroOrMany[str]) -> "Pf":
        return cls.construct(
            name="tags",
            filters={
                cls.DEFAULT: tag,
            },
        )


class ApiVersionTests(unittest.TestCase):
    def test_api_version_string_for_default_fails(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "Cannot convert `DEFAULT` to a valid version string."
        ):
            ApiVersion.DEFAULT.version

    def test_api_version_string_succeeds(self) -> None:
        self.assertEqual("2022-01-01", ApiVersion.VER_2022_01_01.version)


class ApiRequestorTests(unittest.TestCase):
    http_client: ClassVar[HttpClient]
    logger: ClassVar[Logger]
    api_requestor: ClassVar[ApiRequestor]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.http_client = RequestsHttpClient()
        cls.logger = DefaultLogger(name="test-logger")
        cls.api_requestor = ApiRequestor(http_client=cls.http_client, logger=cls.logger)

    def assertAbsoluteUriListEqual(
        self,
        expected: List[str],
        actual: List[Tuple[Any, str, Any, Any]],
    ) -> None:
        results = [
            self.api_requestor.build_uri(
                base_uri=a[0],
                path=a[1],
                path_params=a[2],
                query_params=a[3],
            )
            for a in actual
        ]
        self.assertListEqual(expected, results)

    def test_building_absolute_uri_fails(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "Path requires `hello` parameter, but none is given."
        ):
            self.api_requestor.build_uri(
                base_uri="https://abc.xyz/", path="/{hello}/there/", path_params=None
            )

        with self.assertRaisesRegex(
            ValueError, "Path requires `other` parameter, but none is given."
        ):
            self.api_requestor.build_uri(
                base_uri="https://abc.xyz/",
                path="{param}/hello/{other}/missing/{yeah}",
                path_params={"param": "good_param"},
            )

        with self.assertRaisesRegex(
            ValueError,
            "Path specifies an invalid path parameter. Parameter name must not be empty.",
        ):
            self.api_requestor.build_uri(base_uri="https://abc.xyz/", path="/{}", path_params=None)

    def test_building_absolute_uri_from_api_base_succeeds(self) -> None:
        absolute_uris = [
            (ApiBaseUri.LOCAL, "", None, None),
            (ApiBaseUri.LOCAL, "/{id}", {"id": "12"}, None),
            (ApiBaseUri.PRODUCTION, "/some/simple/path", None, None),
            (
                ApiBaseUri.PRODUCTION,
                "/some/more/{complex}/id/{path}",
                {"complex": "abc_123", "path": "_292_"},
                {"query_a": "1234", "query_b": "abcdefghi_klmo", "12": "34"},
            ),
        ]
        expected = [
            "http://localhost:8000/v2",
            "http://localhost:8000/v2/12",
            "https://api.caplena.com/v2/some/simple/path",
            "https://api.caplena.com/v2/some/more/abc_123/id/_292_?query_a=1234&query_b=abcdefghi_klmo&12=34",
        ]

        self.assertAbsoluteUriListEqual(expected, absolute_uris)

    def test_building_absolute_uri_from_string_succeeds(self) -> None:
        absolute_uris = [
            ("http://abc.de", "", None, None),
            ("https://abc.xyz", "/{id}", {"id": "some_param"}, None),
            (
                "https://caplena.com",
                "/some/simple/path",
                None,
                {"query_a": "1234", "query_b": "abcdefghi_klmo", "12": "34"},
            ),
            (
                "https://nest.js",
                "/some/more/{complex}/id/{path}",
                {"complex": "abc_123", "path": "_292_"},
                None,
            ),
        ]
        expected = [
            "http://abc.de",
            "https://abc.xyz/some_param",
            "https://caplena.com/some/simple/path?query_a=1234&query_b=abcdefghi_klmo&12=34",
            "https://nest.js/some/more/abc_123/id/_292_",
        ]

        self.assertAbsoluteUriListEqual(expected, absolute_uris)

    def test_building_absolute_uri_from_path_parameters_succeeds(self) -> None:
        absolute_uris = [
            ("https://abc.xyz", "/hello/there/", None, None),
            ("https://abc.xyz", "{param}/hello/there/", {"param": "good_param"}, None),
            ("https://abc.xyz", "/{xyz}", {"xyz": "other_param"}, None),
            ("https://abc.xyz/", "/{abc}/", {"abc": "double_slash"}, None),
        ]
        expected = [
            "https://abc.xyz/hello/there/",
            "https://abc.xyz/good_param/hello/there/",
            "https://abc.xyz/other_param",
            "https://abc.xyz/double_slash/",
        ]

        self.assertAbsoluteUriListEqual(expected, absolute_uris)

    def test_building_absolute_uri_from_query_parameters_succeeds(self) -> None:
        absolute_uris = [
            ("https://abc.xyz", "/hello/there/", None, None),
            (
                "https://abc.xyz",
                "param/hello/there/",
                None,
                {"q1": "abc_def", "q2": "123", "4": "5"},
            ),
            ("https://abc.xyz", "/xyz", None, {"_123": "Interesting.Choice/ Ok!"}),
            ("https://abc.xyz/", "/abc_", None, {" ! not sure ": "we can even have spaces"}),
        ]
        expected = [
            "https://abc.xyz/hello/there/",
            "https://abc.xyz/param/hello/there/?q1=abc_def&q2=123&4=5",
            "https://abc.xyz/xyz?_123=Interesting.Choice%2F+Ok%21",
            "https://abc.xyz/abc_?+%21+not+sure+=we+can+even+have+spaces",
        ]

        self.assertAbsoluteUriListEqual(expected, absolute_uris)


class ApiFilterTests(unittest.TestCase):
    def test_constructing_filter_succeeds(self) -> None:
        filters = [
            Pf(),
            Pf.created(),
            Pf.created(year__gt=[10, 20], year__lt=[20]),
        ]
        expected = [
            "ApiFilter()",
            "ApiFilter()",
            "ApiFilter((created.year.gt={10,20}) & (created.year.lt={20}))",
        ]

        self.assertListEqual(expected, [str(filt) for filt in filters])

    def test_or_filter_fails(self) -> None:
        with self.assertRaisesRegex(ValueError, "disjunction of already conjuncted filters"):
            (Pf.created(year__lt=[2000, 1990]) & Pf.created(month=20)) | Pf.created(month=1)

        with self.assertRaisesRegex(ValueError, "disjunction of already conjuncted filters"):
            Pf.created(year__lt=[2000, 1990], month=20) | Pf.created(month=1)

        with self.assertRaisesRegex(ValueError, "as there is already a different filter"):
            Pf.created(year__lt=[2000, 1990]) | Pf.last_modified(day=20)

    def test_or_filter_succeeds(self) -> None:
        filters = [
            Pf.created(year__gt=2020) | Pf(),
            Pf() | Pf.created(year__gt=[10, 20], year__lt=[20]),
            Pf.created(year__gt=10) | Pf.created(),
            Pf.created(year__gt=10) | Pf.created(year__lt=[2, 4, 8]) | Pf.created(day=4),
            Pf.created(year__gt=10) | Pf.created(year__gt=None) | Pf.created(year__gt=12),
        ]
        expected = [
            "ApiFilter((created.year.gt={2020}))",
            "ApiFilter((created.year.gt={10,20}) & (created.year.lt={20}))",
            "ApiFilter((created.year.gt={10}))",
            "ApiFilter((created.year.gt={10} | created.year.lt={2,4,8} | created.day={4}))",
            "ApiFilter((created.year.gt={10,12}))",
        ]

        self.assertListEqual(expected, [str(filt) for filt in filters])

    def test_and_filter_succeeds(self) -> None:
        filters = [
            Pf.created(year__gt=2020) & Pf(),
            Pf() & Pf.created(year__gt=[10, 20], year__lt=[20]),
            Pf.created(year__gt=10) & Pf.created(),
            Pf.created(year__gt=10) & Pf.created(year__lt=[2, 4, 8]) & Pf.created(day=4),
            Pf.created(year__gt=10) & Pf.created(year__gt=None) & Pf.created(year__gt=12),
            Pf.created(year__gt=[10, 20, 30], month=[1, 2, 3], day=[4, 5, 6])
            & Pf.last_modified(year=2021, month=[4, 5, 6], day=None)  # noqa: W503
            & Pf.created(year__lt=[120, 130], year__gt=[1000, 2000]),  # noqa: W503
        ]
        expected = [
            "ApiFilter((created.year.gt={2020}))",
            "ApiFilter((created.year.gt={10,20}) & (created.year.lt={20}))",
            "ApiFilter((created.year.gt={10}))",
            "ApiFilter((created.year.gt={10}) & (created.year.lt={2,4,8}) & (created.day={4}))",
            "ApiFilter((created.year.gt={10}) & (created.year.gt={12}))",
            "ApiFilter((created.year.gt={10,20,30}) & (created.month={1,2,3}) & (created.day={4,5,6}) & (created.year.gt={1000,2000}) "
            "& (created.year.lt={120,130}) & (last_modified.year={2021}) & (last_modified.month={4,5,6}))",
        ]

        self.assertListEqual(expected, [str(filt) for filt in filters])

    def test_default_filter_succeeds(self) -> None:
        filters = [
            Pf.tags("one-tag"),
            Pf.tags(["one-tag", "other-tag"]),
            Pf.tags(["one-tag", "other-tag"]) | Pf.tags(None) | Pf.tags("third-tag"),
        ]
        expected = [
            "ApiFilter((tags={one-tag}))",
            "ApiFilter((tags={one-tag,other-tag}))",
            "ApiFilter((tags={one-tag,other-tag,third-tag}))",
        ]

        self.assertListEqual(expected, [str(filt) for filt in filters])


class ApiFilterQueryParamTests(unittest.TestCase):
    def test_encoding_query_parameters_succeeds(self) -> None:
        filters = [
            Pf.created(year=[2020, 2021, 2022], year__gt=[20, 30, 40], day=[10, 20, 30])
            & (Pf.tags("a") | Pf.tags("b") | Pf.tags("c")),  # noqa: W503
            Pf.created(year__gt=2020, year__lt=2040)
            & Pf.last_modified(day=None, month=12)  # noqa: W503
            & Pf.tags(["a", "b", "c"]),  # noqa: W503
            Pf(),
            Pf.created(),
        ]
        expected = [
            {
                "created": "year:2020,year:2021,year:2022;year.gt:20,year.gt:30,year.gt:40;day:10,day:20,day:30",
                "tags": "a,b,c",
            },
            {
                "created": "year.gt:2020;year.lt:2040",
                "last_modified": "month:12",
                "tags": "a,b,c",
            },
            {},
            {},
        ]

        self.assertListEqual(expected, [filt.to_query_params() for filt in filters])

    def test_encoding_datetime_succeeds(self) -> None:
        filt = Pf.last_modified(
            gte=[
                datetime(2022, 1, 1, tzinfo=timezone.utc),
                datetime(2000, 3, 31, 18, tzinfo=timezone.utc),
            ]
        ) & Pf.last_modified(
            gte=datetime(2021, 5, 17, 12, 10, 50, 500000, tzinfo=timezone.utc)
        )  # noqa: W503
        expected = {
            "last_modified": "gte:2022-01-01T00\\:00\\:00.000Z,gte:2000-03-31T18\\:00\\:00.000Z;gte:2021-05-17T12\\:10\\:50.500Z"
        }

        self.assertEqual(expected, filt.to_query_params())

    def test_escaping_special_characters_succeeds(self) -> None:
        filt = (
            Pf.tags("abc cdef \\ \n ghj xyz") | Pf.tags("th:is;is;just,a,ve:ry;lo:ng;tag")
        ) & Pf.tags("just\\some\\\nmany\\\\\\backslashes\\n")
        expected = {
            "tags": "abc cdef \\\\ \n ghj xyz,th\\:is\\;is\\;just\\,a\\,ve\\:ry\\;lo\\:ng\\;tag;just\\\\some\\\\\nmany\\\\\\\\\\\\backslashes\\\\n"
        }

        self.assertEqual(expected, filt.to_query_params())



from caplena.endpoints.projects_endpoint import (
    ListedProject,
    ProjectDetail,
    Row,
    RowsAppend,
)

__all__ = [
    "ProjectDetail",
    "ListedProject",
    "RowsAppend",
    "Row",
]



from typing import Any, Iterable, List, MutableSequence, TypeVar, Union, overload

T = TypeVar("T")


class CaplenaList(MutableSequence[T]):
    """A custom sequence list, restricting certain modifications of sequence items."""

    _values: List[T]

    _can_replace: bool
    _can_append: bool
    _can_remove: bool

    def __init__(
        self,
        values: Iterable[T] = [],
        *,
        can_replace: bool = False,
        can_append: bool = False,
        can_remove: bool = False,
    ) -> None:
        self._values = list(values)

        # TODO: implement removing, replacing and appending logic in `BaseObject` diff logic.
        if can_replace or can_append or can_remove:
            raise ValueError(
                "Replacing, appending and removing is currently not supported. HINT: To support it, "
                "please implement the respective diff. logic in `BaseObject`."
            )
        self._can_replace = can_replace
        self._can_append = can_append
        self._can_remove = can_remove

    def insert(self, index: int, value: T) -> None:
        if self._can_append is False:
            raise ValueError(
                f"Error appending item at index {index}. HINT: You cannot append any items to this list."
            )

        self._values.insert(index, value)

    def __str__(self) -> str:
        return str(self._values)

    def __repr__(self) -> str:
        if len(self._values) > 5:
            reprs = [value.__repr__() for value in self._values[:5]]
            return f"[{', '.join(reprs)}, ...]"
        else:
            return self._values.__repr__()

    def __len__(self) -> int:
        return len(self._values)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CaplenaList):
            return False

        return self._values == other._values  # pyright: reportUnknownMemberType=false

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[T, MutableSequence[T]]:
        return self._values[index]

    @overload
    def __setitem__(self, index: int, value: T) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[T]) -> None:
        ...

    def __setitem__(self, index: Union[int, slice], value: Union[T, Iterable[T]]) -> None:
        if isinstance(index, slice):
            raise ValueError("Using slices to set list items is not supported.")

        if self._can_replace is False:
            raise ValueError(
                f"Error setting item at index {index}. HINT: You cannot replace any items from this list. If you "
                "want to modify an object, consider updating its properties instead of replacing the object itself."
            )

        self._values[index] = value  # type: ignore

    @overload
    def __delitem__(self, index: int) -> None:
        ...

    @overload
    def __delitem__(self, index: slice) -> None:
        ...

    def __delitem__(self, index: Union[int, slice]) -> None:
        if self._can_remove is False:
            raise ValueError(
                f"Error deleting item at index {index}. HINT: You cannot remove any items from this list."
            )

        del self._values[index]



import json
from typing import Any, Dict, Optional


class HttpResponse:
    _json: Optional[Dict[str, Any]]

    @property
    def json(self) -> Optional[Dict[str, Any]]:
        if hasattr(self, "_json"):
            return self._json
        elif self.text:
            self._json = json.loads(self.text)
            return self._json
        else:
            return None

    def __init__(
        self,
        *,
        status_code: int,
        reason: str,
        text: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.status_code = status_code
        self.reason = reason
        self.text = text
        self.headers = headers

    def __str__(self) -> str:
        concat_text = (
            self.text[:50] + "..." if self.text and len(self.text) > 50 else str(self.text)
        )
        return f"HttpResponse(status_code={self.status_code}, reason='{self.reason}', text='{concat_text}')"



from enum import Enum
from json import dumps
from typing import Any, ClassVar, Dict, Iterable, List, Optional, Union

from caplena.http.http_response import HttpResponse
from caplena.http.json_encoder import JsonDateEncoder
from caplena.logging.default_logger import DefaultLogger
from caplena.logging.logger import Logger


class HttpMethod(Enum):
    GET = 0
    POST = 1
    PUT = 2
    PATCH = 3
    DELETE = 4
    HEAD = 5

    @property
    def method(self) -> str:
        return self.name.lower()

    def __str__(self) -> str:
        return self.name


class HttpRetry:
    DEFAULT_MAX_RETRIES: ClassVar[int] = 0
    DEFAULT_BACKOFF_FACTOR: ClassVar[float] = 2
    DEFAULT_STATUS_CODES_TO_RETRY: ClassVar[Iterable[int]] = frozenset(
        {408, 409, 413, 429, 500, 502, 503, 504}
    )
    DEFAULT_ALLOWED_METHOD = frozenset({HttpMethod.GET, HttpMethod.PUT, HttpMethod.HEAD})

    def __init__(
        self,
        *,
        max_retries: int = DEFAULT_MAX_RETRIES,
        backoff_factor: float = DEFAULT_BACKOFF_FACTOR,
        retry_status_codes: Iterable[int] = DEFAULT_STATUS_CODES_TO_RETRY,
        retry_methods: Iterable[HttpMethod] = DEFAULT_ALLOWED_METHOD,
    ):
        self.max_retries = max_retries
        self.backoff_factor = backoff_factor
        self.retry_status_codes = retry_status_codes
        self.retry_methods = retry_methods


class HttpClient:
    DEFAULT_TIMEOUT: ClassVar[int] = 120
    DEFAULT_RETRY: ClassVar[HttpRetry] = HttpRetry()
    DEFAULT_LOGGER: ClassVar[Logger] = DefaultLogger("http[shared]")
    DEFAULT_ENCODER: ClassVar[JsonDateEncoder] = JsonDateEncoder()

    @property
    def identifier(self) -> str:
        raise NotImplementedError("HttpClient subclasses must provide a `identifier` property.")

    def __init__(
        self,
        *,
        timeout: int = DEFAULT_TIMEOUT,
        retry: HttpRetry = DEFAULT_RETRY,
        logger: Logger = DEFAULT_LOGGER,
        encoder: JsonDateEncoder = DEFAULT_ENCODER,
    ):
        self.timeout = timeout
        self.retry = retry
        self.logger = logger
        self.encoder = encoder

    def request(
        self,
        uri: str,
        *,
        method: HttpMethod = HttpMethod.GET,
        headers: Optional[Dict[str, str]] = None,
        json: Optional[Union[Dict[str, Any], List[Any]]] = None,
        timeout: Optional[int] = None,
        retry: Optional[HttpRetry] = None,
    ) -> HttpResponse:
        timeout = self.get_timeout(timeout)
        retry = self.get_retry(retry)
        self.logger.info("Sending request to Caplena API", method=str(method), uri=uri)

        data = None
        headers = headers if headers is not None else {}
        if json is not None:
            headers["content-type"] = "application/json"
            data = dumps(json, cls=JsonDateEncoder)
            self.logger.debug("Sending request to Caplena API", data=data)

        # TODO: handle retry here
        response = self.request_raw(
            uri=uri,
            method=method,
            timeout=timeout,
            headers=headers,
            data=data,
        )
        self.logger.debug(
            "Received response from server",
            status_code=str(response.status_code),
            text=str(response.text),
        )

        return response

    def request_raw(
        self,
        uri: str,
        *,
        method: HttpMethod,
        timeout: int,
        headers: Dict[str, str],
        data: Optional[str] = None,
    ) -> HttpResponse:
        raise NotImplementedError("HttpClient subclasses must implement `request_raw`.")

    def get_timeout(self, timeout: Optional[int] = None) -> int:
        return timeout if timeout is not None else self.timeout

    def get_retry(self, retry: Optional[HttpRetry] = None) -> HttpRetry:
        return retry if retry is not None else self.retry



from enum import Enum


class LoggingLevel(Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3

    @property
    def level(self) -> int:
        return self.value


class Logger:
    @property
    def name(self) -> str:
        return self._name

    @property
    def logging_level(self) -> LoggingLevel:
        return self._logging_level

    def __init__(self, name: str, logging_level: LoggingLevel = LoggingLevel.WARNING):
        self._name = name
        self._logging_level = logging_level

    def log(self, msg: str, *, level: LoggingLevel, **extra: str) -> None:
        raise NotImplementedError("Logger subclasses must implement `log`.")

    def debug(self, msg: str, **extra: str) -> None:
        self.log(msg, level=LoggingLevel.DEBUG, **extra)

    def info(self, msg: str, **extra: str) -> None:
        self.log(msg, level=LoggingLevel.INFO, **extra)

    def warning(self, msg: str, **extra: str) -> None:
        self.log(msg, level=LoggingLevel.WARNING, **extra)

    def error(self, msg: str, **extra: str) -> None:
        self.log(msg, level=LoggingLevel.ERROR, **extra)
